﻿using System;
using System.Management;
using System.Runtime.InteropServices;
using System.Text;
using Microsoft.Win32.SafeHandles;
using System.IO;
using System.Linq;
using System.Collections.Generic;

namespace USBHelper
{
    public class USBPhysicalDrive
    {
        const uint GENERIC_READ = 0x80000000;
        const uint GENERIC_WRITE = 0x40000000;
        const uint FILE_SHARE_READ = 0x00000001;
        const uint FILE_SHARE_WRITE = 0x00000002;
        const uint OPEN_EXISTING = 3;
        const uint IOCTL_STORAGE_EJECT_MEDIA = 0x2D4808;

        [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        static extern SafeFileHandle CreateFile(
            string lpFileName,
            uint dwDesiredAccess,
            uint dwShareMode,
            IntPtr lpSecurityAttributes,
            uint dwCreationDisposition,
            uint dwFlagsAndAttributes,
            IntPtr hTemplateFile);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool DeviceIoControl(
            SafeFileHandle hDevice,
            uint dwIoControlCode,
            IntPtr lpInBuffer,
            uint nInBufferSize,
            IntPtr lpOutBuffer,
            uint nOutBufferSize,
            out uint lpBytesReturned,
            IntPtr lpOverlapped);

        public static string GetPnpDeviceIdFromDriveLetter(string driveLetter)
        {
            string query = string.Format("ASSOCIATORS OF {{Win32_LogicalDisk.DeviceID='{0}'}} WHERE AssocClass=Win32_LogicalDiskToPartition", driveLetter);
            using (ManagementObjectSearcher partitionSearcher = new ManagementObjectSearcher(query))
            {
                foreach (ManagementObject partition in partitionSearcher.Get())
                {
                    string partitionDeviceId = partition["DeviceID"].ToString();
                    string diskQuery = string.Format("ASSOCIATORS OF {{Win32_DiskPartition.DeviceID='{0}'}} WHERE AssocClass=Win32_DiskDriveToDiskPartition", partitionDeviceId);
                    using (ManagementObjectSearcher diskSearcher = new ManagementObjectSearcher(diskQuery))
                    {
                        foreach (ManagementObject diskDrive in diskSearcher.Get())
                        {
                            return diskDrive["PNPDeviceID"].ToString();
                        }
                    }
                }
            }
            return null;
        }

        public static bool EjectPhysicalDrive(string physicalDrivePath)
        {
            using (SafeFileHandle handle = CreateFile(
                physicalDrivePath,
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                IntPtr.Zero,
                OPEN_EXISTING,
                0,
                IntPtr.Zero))
            {
                if (handle.IsInvalid)
                {
                    int err = Marshal.GetLastWin32Error();
                    Console.WriteLine("Failed to open handle: error " + err);
                    return false;
                }

                uint bytesReturned;
                bool result = DeviceIoControl(
                    handle,
                    IOCTL_STORAGE_EJECT_MEDIA,
                    IntPtr.Zero, 0,
                    IntPtr.Zero, 0,
                    out bytesReturned,
                    IntPtr.Zero);

                if (!result)
                {
                    int err = Marshal.GetLastWin32Error();
                    Console.WriteLine("Eject failed: error " + err);
                }

                return result;
            }
        }

        public static string GetPhysicalDriveForLetter(string driveLetter)
        {
            if (!driveLetter.EndsWith(":"))
                driveLetter += ":";

            string query = "ASSOCIATORS OF {Win32_LogicalDisk.DeviceID='" + driveLetter +
                           "'} WHERE AssocClass = Win32_LogicalDiskToPartition";

            using (var searcher = new System.Management.ManagementObjectSearcher(query))
            {
                foreach (var partition in searcher.Get())
                {
                    string partitionId = partition["DeviceID"].ToString();
                    string diskQuery = "ASSOCIATORS OF {Win32_DiskPartition.DeviceID='" +
                                       partitionId + "'} WHERE AssocClass = Win32_DiskDriveToDiskPartition";

                    using (var diskSearcher = new System.Management.ManagementObjectSearcher(diskQuery))
                    {
                        foreach (var disk in diskSearcher.Get())
                        {
                            string deviceId = disk["DeviceID"].ToString(); // like \\.\PhysicalDrive2
                            return deviceId;
                        }
                    }
                }
            }

            return null;
        }

        public static string GetPhysicalDrivePath(string targetDriveLetter)
        {
            try
            {
                var searcher = new ManagementObjectSearcher("SELECT * FROM Win32_LogicalDiskToPartition");

                foreach (ManagementObject mo in searcher.Get())
                {
                    string antecedent = mo["Antecedent"].ToString(); // Partition
                    string dependent = mo["Dependent"].ToString();   // Drive letter

                    string partition = ParseWmiValue(antecedent);
                    string driveLetter = ParseWmiValue(dependent);

                    if (string.Equals(driveLetter, targetDriveLetter, StringComparison.OrdinalIgnoreCase))
                    {
                        int diskNumber = GetDiskNumber(partition);
                        return @"\\.\PhysicalDrive" + diskNumber;
                    }
                }
            }
            catch
            {
                // You could log or handle this more gracefully
            }

            return null; // not found
        }

        private static string ParseWmiValue(string input)
        {
            int start = input.IndexOf("=\"") + 2;
            int end = input.IndexOf("\"", start);
            return input.Substring(start, end - start);
        }

        private static int GetDiskNumber(string partition)
        {
            int diskIndex = partition.IndexOf("Disk #") + 6;
            int commaIndex = partition.IndexOf(",", diskIndex);
            string diskNumStr = partition.Substring(diskIndex, commaIndex - diskIndex);
            return int.Parse(diskNumStr);
        }

    }
}
